#! /usr/bin/python
# Copyright (C) 2006 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""A HTTP server that translates http into https

Usage:

python InteProxy.py
"""

import sys
import optparse
import traceback
import time
import urlparse
import urllib2
import BaseHTTPServer
import threading
import Queue
import socket

try:
    import pywin.dialogs.login
    windowsystem = "pywin"
except ImportError:
    windowsystem = "crt"


def getpassword_pywin(title):
    return pywin.dialogs.login.GetLogin(title)

def getpassword_crt(title):
    import getpass
    try:
        user = raw_input(title)
        password = getpass.getpass()
    except IOError:
        import traceback
        traceback.print_exception()
        user = password = None
    return user, password

getpassword = globals()["getpassword_" + windowsystem]

# the central queue for the communication between master and worker
# thread.
the_queue = Queue.Queue(0)

# password cache
pw_cache = {}

def worker_thread():
    """The worker thread

    This function takes requests out of the queue and handles them.
    """
    while 1:
        request = the_queue.get()
        request()


class InteProxyHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    # Whether the SHUTDOWN method is allowed.  If allowed, anybody who can
    # connect to the server can shutdown the server.
    allow_shutdown = False

    def do_SHUTDOWN(self):
        """Shutdown the server if self.allow_shutdown is True (default False).

        To shutdown the server, this method calls the server's
        server_close method.
        """
        if self.allow_shutdown:
            self.server.server_close()
            self.send_response(200)
            # somewhat ugly hack: connect to this server, to unblock the
            # accept call.
            s = socket.socket(socket.AF_INET)
            s.connect(("localhost", self.server.server_port))
            s.close()
        else:
            self.send_error(400)

    def do_GET(self):
        """Handle a GET request"""
        self.handle_proxy_request("GET")

    def do_POST(self):
        """Handle a POST request"""
        self.handle_proxy_request("POST")

    def handle_proxy_request(self, method):
        """Handle a request that needs to be 'proxied'

        Method should be either 'POST' or 'GET'.  Both methods are
        handled in the same way.  The real host to connect to is
        determined from the url: The path component starts with the real
        hostname.  The method strips that hostname from the path and
        connects to the new host.  All headers from the HTTP request and
        any extra data are passed through to the remote host and the
        reponse is likewise simply copied through.  The only exception
        is the Host: header from the original request.  It would be
        wrong, so it is removed and a new one is generated pointing to
        the real remote host.
        """
        netloc, url = self.convert_url(self.path)

        #
        # Create a http request for the real location
        #
        request = urllib2.Request("https://%s/%s" % (netloc, url))
        for header, value in self.headers.items():
            if header.lower() == "host":
                # the host header will be set by httplib and it should
                # not be set twice, so we omit it.  The value we
                # received from the client would have been wrong anyway
                # as it refers to the host this proxy is running on.
                continue
            elif header.lower() in ("content-length", "content-type"):
                # When a post request is redirected with the 302 code,
                # urllib2 will do a GET request for the new url even
                # when the original request was POST.  This does not
                # conform to RFC 2616, but it's apparently what many
                # clients do and some servers expect (e.g. a POST
                # request to http://intevation.de/roundup/kolab/ is
                # redirected to
                # https://intevation.de/roundup/kolab/issue<num> and the
                # server expects a GET request for that.).  However,
                # when doing the new request urllib2 copies all headers
                # except for those added with add_unredirected_header.
                # The content-length and content-type headers are only
                # needed for POST request so we have to make sure
                # they're not send when redirection occurs.
                request.add_unredirected_header(header, value)
            else:
                request.add_header(header, value)

        # If any data is associated with the request pass it through
        length = int(self.headers.getheader("Content-Length", "0"))
        if length:
            # FIXME: check whether length bytes were read
            request.add_data(self.rfile.read(length))

        #
        # Retrieve the url described by the request and pass everything
        # through to the client.
        #
        try:
            response = urllib2.urlopen(request)
        except urllib2.HTTPError, err:
            # a HTTPError is a valid http response object, so we can
            # treat it like a normal response.
            response = err
        self.send_response(response.code, response.msg)
        for header, value in response.info().items():
            self.send_header(header, value)
        self.end_headers()

        transfer_encoding = response.info().get("Transfer-encoding")
        self.transfer_data(response.read, self.wfile.write,
                           chunked = (transfer_encoding == "chunked"))

    def transfer_data(self, read, write, length=None, chunked=False):
        """Transfer data from one 'file' to another in chunks

        The files are given by their read and write methods so it
        doesn't have to be a file.  The read parameter must be callable
        with an integer argument indicating the maximum number of bytes
        to read and the write parameter must be callable with a string.
        If the parameter chunked is true, the method uses the 'chunked'
        transfer encoding when writing the data.
        """
        max_chunk_size = 4096
        while 1:
            if length is not None:
                chunk_size = min(length, max_chunk_size)
            else:
                chunk_size = max_chunk_size
            chunk = read(chunk_size)
            if not chunk:
                break
            if length is not None:
                length -= len(chunk)
            if chunked:
                write("%x\r\n" % len(chunk))
            write(chunk)
            if chunked:
                write("\r\n")
        if chunked:
            write("0\r\n\r\n")

    def convert_url(self, url):
        """Extract the real host from the url and return host and new url

        The URL parameter is the part of the URL from the GET or POST
        request which is actually only the path, query and fragment in
        the terminology of the urlparse module.  For the InteProxy, the
        first component in the path is the actual host to connect to.
        This method extracts that part and creates a new path with that
        part removed.  The return value of this method is the tuple
        (real_host, new_url) where new_url is the combined new path and
        the original query and fragment parts.
        """
        scheme, netloc, path, query, fragment = urlparse.urlsplit(url)

        # XXX: This is hardcoded syntax for deegree OWS Proxy
        # Eventually this should be moved to a method of its own
        # and anyway this is just a dirty solution since passwords
        # are not really managed.
        if pw_cache.has_key(path):
            user, password = pw_cache[path]
        else:
            title = "Username for %s: " % path
            user, password = getpassword(title)
        # TODO: Building the query this way is not safe.
        query += "&user=%s&password=%s" % (user, password)
        pw_cache[path] = (user, password)

        # both scheme and netloc must be empty strings because the url
        # we work with is the url on the first line of a HTTP request
        # which doesn't include them
        assert scheme == ""
        assert netloc == ""

        split_path = path.split("/")
        if not split_path[0]:
            del split_path[0]
        real_netloc = split_path.pop(0)
        real_path = "/" + "/".join(split_path)
        return real_netloc, urlparse.urlunsplit(("", "", real_path, query,
                                                 fragment))

    def log_exception(self, exc_info):
        """Log an exception

        This method produces a log message via self.log_message and then
        prints a the exception given by exc_info, which should be the
        return value of sys.exc_info, to stderr.  By default log_message
        also writes to stderr, so both the message and the traceback
        should go to the same place.
        """
        self.log_message("Exception for previous 500 code\n")
        traceback.print_exception(*(exc_info + (None, sys.stderr)))
        sys.stderr.flush()


# same as the BaseHTTPRequestHandler method, but as a standalone function:
def log_date_time_string():
    """Return the current time formatted for logging."""
    now = time.time()
    year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
    s = "%02d/%3s/%04d %02d:%02d:%02d" % (
        day, BaseHTTPServer.BaseHTTPRequestHandler.monthname[month],
        year, hh, mm, ss)
    return s


class MasterWorkerServer(BaseHTTPServer.HTTPServer):

    do_shutdown = False

    def serve_forever(self):
        """Handle requests until self.do_shutdown is True."""
        while not self.do_shutdown:
            self.handle_request()

    def server_close(self):
        """Shutdown the server.

        Usually this is called from a thread while another thread
        executes the server_forever method.  In that case the other
        thread is most likely blocked while listening for new
        connections.  After this method returns, the caller will have to
        do something to unblock the other thread, such as creating a TCP
        connection to the server.
        """
        BaseHTTPServer.HTTPServer.server_close(self)
        self.do_shutdown = True

    def process_request(self, request, client_address):
        """Put the request into the queue to be handled by the worker thread
        """
        def process_in_worker():
            try:
                self.finish_request(request, client_address)
                self.close_request(request)
            except:
                self.handle_error(request, client_address)
                self.close_request(request)
        sys.stderr.write("[%s] queue contains %d items\n" %
                         (log_date_time_string(), the_queue.qsize()))
        the_queue.put(process_in_worker)

    def handle_error(self, request, client_address):
        """Override to integrate the error reporting better with the other logs
        """
        sys.stderr.write("[%s] Exception while handling request from %r\n"
                         % (log_date_time_string(), client_address))
        traceback.print_exception(*(sys.exc_info() + (None, sys.stderr)))
        sys.stderr.flush()


class IntePasswordManager(urllib2.HTTPPasswordMgr):

    def find_user_password(self, realm, authuri):
        user, password \
              = urllib2.HTTPPasswordMgr.find_user_password(self, realm, authuri)
        if user is None:
            # no password known yet.  Ask the user
            authuri = "https://" + "".join(self.reduce_uri(authuri))
            title = "Username for %r realm %r: " % (authuri, realm)
            user, password = getpassword(title)
        return user, password

def setup_urllib2():
    auth_handler = urllib2.HTTPBasicAuthHandler(IntePasswordManager())
    httphandler = urllib2.HTTPHandler()
    opener = urllib2.build_opener(auth_handler, httphandler)
    urllib2.install_opener(opener)


def run_server(HandlerClass = InteProxyHTTPRequestHandler,
               ServerClass = MasterWorkerServer):
    """Run the InteProxy server"""
    parser = optparse.OptionParser()
    parser.set_defaults(port=64609)
    parser.add_option("--logfile")
    parser.add_option("--allow-shutdown", action="store_true")
    parser.add_option("--port", type="int")
    opts, rest = parser.parse_args()

    HandlerClass.allow_shutdown = opts.allow_shutdown

    setup_urllib2()

    server_address = ('localhost', opts.port)

    if opts.logfile:
        sys.stderr = open(opts.logfile, "a", 0)

    sys.stderr.write("[%s] server starting up\n" % log_date_time_string())

    httpd = ServerClass(server_address, HandlerClass)

    sys.stderr.write("[%s] starting worker thread\n" % log_date_time_string())
    worker = threading.Thread(target = worker_thread)
    worker.setDaemon(1)
    worker.start()

    print "Serving HTTP on port", opts.port, "..."
    sys.stderr.write("[%s] serving HTTP on port %d\n"
                     % (log_date_time_string(), opts.port))
    httpd.serve_forever()
    logmessage = "[%s] server stopped" % (log_date_time_string(),)
    print logmessage
    sys.stderr.write(logmessage + "\n")


if __name__ == '__main__':
    run_server()
